#include "canvas.h"
#include "mem.h"
#include <stdio.h>

int initCanvas(Canvas *canvas, uint16_t width, uint16_t height, uint8_t *data)
 {
  assert(canvas);
  assert(width > 0);
  assert(height > 0);
  assert(data);
  
  canvas->width = width;
  canvas->height = height;
  canvas->pitch = width*3;
  canvas->data = data;
  canvas->points_begin = canvas->points_end = canvas->points_alloc_end = 0;
  DRInit(&canvas->dirt);
  return 1;
 }

void uninitCanvas(Canvas *canvas)
 {
  assert(canvas);
  
  if(canvas->points_begin)
   free(canvas->points_begin);
 }

Canvas *createCanvas(uint16_t  width, uint16_t height)
 {
  Canvas *canvas = malloc(sizeof(Canvas)+width*height*3);
  
  if(!canvas)
   return 0;
  
  if(!initCanvas(canvas, width, height, (uint8_t *)canvas+sizeof(Canvas)))
   {
    free(canvas);
    return 0;
   }
  
  return canvas;
 }

void destroyCanvas(Canvas *canvas)
 {
  assert(canvas);
  uninitCanvas(canvas);
  free(canvas);
 }

void drawCanvasPoint(Canvas *canvas, CanvasPoint p, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
 {
  signed int ix = p.x, iy = p.y;
  float fx=p.x-ix, ifx=1.0f-fx, fy=p.y-iy, ify=1.0f-fy;
  
  if(ix >= -1 && iy >= -1 && ix < canvas->width && iy < canvas->height)
   {
    uint32_t pitch = canvas->pitch;
    uint8_t *p = canvas->data+ix*3+iy*pitch;
    
    if(ix == -1)
     {
      if(iy == -1)
       {
        /*  ..  *
         *  .o  */
        writeAlphaPixel(p+pitch+3, r, g, b, fx*fy*a);
       }
      else if(iy == canvas->height-1)
       {
        /*  .o  *
         *  ..  */
        writeAlphaPixel(p+3, r, g, b, fx*ify*a);
       }
      else
       {
        /*  .o  *
         *  .o  */
        p += 3;
        writeAlphaPixel(p, r, g, b, fx*ify*a);
        writeAlphaPixel(p+pitch, r, g, b, fx*fy*a);
       }
     }
    else if(ix == canvas->width-1)
     {
      if(iy == -1)
       {
        /*  ..  *
         *  o.  */
        writeAlphaPixel(p+pitch, r, g, b, ifx*fy*a);
       }
      else if(iy == canvas->height-1)
       {
        /*  o.  *
         *  ..  */
        writeAlphaPixel(p, r, g, b, ifx*ify*a);
       }
      else
       {
        /*  o.  *
         *  o.  */
        writeAlphaPixel(p, r, g, b, ifx*ify*a);
        writeAlphaPixel(p+pitch, r, g, b, ifx*fy*a);
       }
     }
    else
     {
      if(iy == -1)
       {
        /*  ..  *
         *  oo  */
        p += pitch;
        writeAlphaPixel(p, r, g, b, ifx*fy*a);
        writeAlphaPixel(p+3, r, g, b, fx*fy*a);
       }
      else if(iy == canvas->height-1)
       {
        /*  oo  *
         *  ..  */
        writeAlphaPixel(p, r, g, b, ifx*ify*a);
        writeAlphaPixel(p+3, r, g, b, fx*ify*a);
       }
      else
       {
        /*  oo  *
         *  oo  */
        writeAlphaPixel(p, r, g, b, ifx*ify*a);
        writeAlphaPixel(p+3, r, g, b, fx*ify*a);
        p += pitch;
        writeAlphaPixel(p, r, g, b, ifx*fy*a);
        writeAlphaPixel(p+3, r, g, b, fx*fy*a);
       }
     }
   }
 }

float getCanvasColour(const Canvas *canvas, CanvasPoint p, float *pixel)
 {
  /* I'm not confident the edge cases here are correct. Need to test. */
  
  signed int ix = p.x, iy = p.y;
  float fx=p.x-ix, ifx=1.0f-fx, fy=p.y-iy, ify=1.0f-fy;
  
  if(ix >= -1 && iy >= -1 && ix < canvas->width && iy < canvas->height)
   {
    uint32_t pitch = canvas->pitch;
    uint8_t *p = canvas->data+ix*3+iy*pitch;
    
    if(ix == -1)
     {
      if(iy == -1)
       {
        /*  ..  *
         *  .o  */
        p += pitch+3;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p;
        return fx*fy;
       }
      else if(iy == canvas->height-1)
       {
        /*  .o  *
         *  ..  */
        p += 3;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p;
        return fx*ify;
       }
      else
       {
        /*  .o  *
         *  .o  */
        uint8_t *p2 = (p+=3)+pitch;
        *pixel = fy**p2-fy**p+*p; ++pixel; ++p; ++p2;
        *pixel = fy**p2-fy**p+*p; ++pixel; ++p; ++p2;
        *pixel = fy**p2-fy**p+*p;
        return fx;
       }
     }
    else if(ix == canvas->width-1)
     {
      if(iy == -1)
       {
        /*  ..  *
         *  o.  */
        p += pitch;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p;
        return ifx*fy;
       }
      else if(iy == canvas->height-1)
       {
        /*  o.  *
         *  ..  */
        *pixel = *p; ++pixel; ++p;
        *pixel = *p; ++pixel; ++p;
        *pixel = *p;
        return ifx*ify;
       }
      else
       {
        /*  o.  *
         *  o.  */
        uint8_t *p2 = p+pitch;
        *pixel = fy**p2*-fy**p+*p; ++pixel; ++p; ++p2;
        *pixel = fy**p2*-fy**p+*p; ++pixel; ++p; ++p2;
        *pixel = fy**p2*-fy**p+*p;
        return ifx;
       }
     }
    else
     {
      if(iy == -1)
       {
        /*  ..  *
         *  oo  */
        uint8_t *p2 = (p+=pitch)+3;
        *pixel = fx**p2*-fx**p+*p; ++pixel; ++p; ++p2;
        *pixel = fx**p2*-fx**p+*p; ++pixel; ++p; ++p2;
        *pixel = fx**p2*-fx**p+*p;
        return fy;
       }
      else if(iy == canvas->height-1)
       {
        /*  oo  *
         *  ..  */
        uint8_t *p2 = p+3;
        *pixel = fx**p2*-fx**p+*p; ++pixel; ++p; ++p2;
        *pixel = fx**p2*-fx**p+*p; ++pixel; ++p; ++p2;
        *pixel = fx**p2*-fx**p+*p;
        return ify;
       }
      else
       {
        /*  oo  *
         *  oo  */
        uint8_t *p2 = p+3, *p3=p+pitch, *p4=p3+3;
        *pixel = *p*ifx*ify+*p2*fx*ify+*p3*ifx*fy+*p4*fx*fy; ++pixel; ++p; ++p2; ++p3; ++p4;
        *pixel = *p*ifx*ify+*p2*fx*ify+*p3*ifx*fy+*p4*fx*fy; ++pixel; ++p; ++p2; ++p3; ++p4;
        *pixel = *p*ifx*ify+*p2*fx*ify+*p3*ifx*fy+*p4*fx*fy;
        return 1.0f;
       }
     }
   }
  
  return 0.0f;
 }

void dumpCanvasPoints(Canvas *canvas, uint8_t r, uint8_t g, uint8_t b, uint8_t a)
 {
  const CanvasPoint *p = canvas->points_begin;
  
  Region region;
  
  if(p != canvas->points_end)
   {
    region.x2 = (region.x1 = p->x) + 2;
    region.y2 = (region.y1 = p->y) + 2;
    
    for(; p != canvas->points_end; ++p)
     {
      int16_t t;
      
      
      if((t = p->x) < region.x1)
       region.x1 = t;
      else if((t += 2) > region.x2)
       region.x2 = t;
      
      if((t = p->y) < region.y1)
       region.y1 = t;
      else if((t += 2) > region.y2)
       region.y2 = t;
      
      drawCanvasPoint(canvas, *p, r, g, b, a);
     }
   }
  
  DRAdd(&canvas->dirt, &region);
  
  clearCanvasPoints(canvas);
 }

void clearCanvasPoints(Canvas *canvas)
 {
  assert(canvas);
  canvas->points_end = canvas->points_begin;
 }

int pushCanvasPoint(Canvas *canvas, CanvasPoint p)
 {
  assert(canvas);
  assert(isfinite(p.x) && isfinite(p.y));
  
  if(canvas->points_end == canvas->points_alloc_end)
   {
    
    if(canvas->points_begin)
     {
      CanvasPoint *new_data = (CanvasPoint *)realloc(canvas->points_begin, sizeof(CanvasPoint)*2*(canvas->points_alloc_end-canvas->points_begin));
      if(!new_data) return 0;
      
      canvas->points_end = new_data+(canvas->points_end-canvas->points_begin);
      canvas->points_alloc_end = new_data+2*(canvas->points_alloc_end-canvas->points_begin);
      canvas->points_begin = new_data;
     }
    else if((canvas->points_begin = (CanvasPoint *)malloc(sizeof(CanvasPoint)*128)))
     {
      canvas->points_end = canvas->points_begin;
      canvas->points_alloc_end = canvas->points_begin+128;
     }
    else return 0;
   }
  
  *(canvas->points_end) = p;
  ++canvas->points_end;
  return 1;
 }

int pushCanvasLine(Canvas *canvas, CanvasPoint start, CanvasPoint end, float *remainder)
 {
  float distance;
  
  assert(remainder);
  assert(isfinite(*remainder));
  assert(*remainder >= 0.0);
  
  end.x -= start.x;
  end.y -= start.y;
  
  distance = sqrt(end.x*end.x+end.y*end.y);
  
  if(distance < *remainder || distance < 0.000001f)
   {
    *remainder -= distance;
    return 1;
   }
  
  end.x /= distance;
  end.y /= distance;
  
  assert(isfinite(end.x) && isfinite(end.y));
  
  start.x += *remainder*end.x;
  start.y += *remainder*end.y;
  
  distance -= *remainder;
  
  while(pushCanvasPoint(canvas, start))
   {
    if(distance < 1.0f)
     {
      *remainder = 1.0f - distance;
      return 1;
     }
    else
     {
      start.x += end.x;
      start.y += end.y;
      distance -= 1.0f;
     }
   }
  
  return 0;
 }
