#include "mem.h"
#include "glcanvas.h"

#include "SDL/SDL.h"
#include <GL/gl.h>

typedef struct
 {
  Canvas canvas;
  unsigned int tw, th;
  unsigned int xremainder, yremainder;
  GLuint *tex;
 } GLCanvas;

#include <math.h>
#include <stdio.h>

Canvas *createGLCanvas(unsigned int width, unsigned int height)
 {
  unsigned int tw = (width+255)/256,
               th = (height+255)/256;
  
  GLCanvas *canvas = malloc(sizeof(GLCanvas)+tw*th*sizeof(GLuint)+width*height*3);
  
  if(!canvas)
   return 0;
  
  if(!initCanvas(&canvas->canvas, width, height, ((unsigned char *)canvas)+(sizeof(GLCanvas)+tw*th*sizeof(GLuint))))
   {
    free(canvas);
    return 0;
   }
  
  canvas->tw = tw;
  canvas->th = th;
  canvas->xremainder = width-(tw-1)*256;
  canvas->yremainder = height-(th-1)*256;
  canvas->tex = (GLuint *)(((unsigned char *)canvas)+sizeof(GLCanvas));
  
  glPushAttrib(GL_TEXTURE_BIT | GL_ENABLE_BIT); // Don't mess up the textures of anyone else.
  glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
  glEnable(GL_TEXTURE_2D);
  
  glGenTextures(tw*th, canvas->tex);
  
  glPixelStorei(GL_UNPACK_ROW_LENGTH, width);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  
  renderGLCanvasMandelbrot(&canvas->canvas);
  
  unsigned int y, x;
  
  for(y = 0; y < th; ++y)
   for(x = 0; x < tw; ++x)
    {
     glBindTexture(GL_TEXTURE_2D, canvas->tex[x+y*tw]);
     
     glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
                  x == tw-1 ? canvas->xremainder : 256,
                  y == th-1 ? canvas->yremainder : 256,
                  0, GL_RGB, GL_UNSIGNED_BYTE,
                  canvas->canvas.data+3*256*x+256*canvas->canvas.pitch*y);
     
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
     glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    }
  
  glPopClientAttrib();
  glPopAttrib();
  
  return &canvas->canvas;
 }

void destroyGLCanvas(Canvas *canvas)
 {
  GLCanvas *glc = (GLCanvas *)canvas;
  glDeleteTextures(glc->tw*glc->th, glc->tex);
  uninitCanvas(canvas);
  free(canvas);
 }

void drawGLCanvas(const Canvas *canvas)
 {
  GLCanvas *glc = (GLCanvas *)canvas;
  
  glPushAttrib(GL_TEXTURE_BIT | GL_ENABLE_BIT | GL_CURRENT_BIT);
  
  glEnable(GL_TEXTURE_2D);
  glColor3ub(255, 255, 255);
  
  GLuint *tex = glc->tex;
  unsigned int tw = glc->tw, th = glc->th;
  unsigned int xrem = glc->xremainder, yrem = glc->yremainder;
  unsigned int y, x;
  
  if(DRIsDirty(&canvas->dirt))
   {
    unsigned char *canvas_data = canvas->data;
    unsigned int pitch = canvas->pitch;
    
    glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
    glPixelStorei(GL_UNPACK_ROW_LENGTH, canvas->width);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
    Region region = {0, 0, canvas->width, canvas->height};
    
    DRClip((DirtyRectangles *)&canvas->dirt, &region);
    DROptimize((DirtyRectangles *)&canvas->dirt);
    
    if(DRIsSaturated(&canvas->dirt))
     {
      for(y = 0; y < th; ++y)
       {
        unsigned char *data = canvas_data+y*256*pitch;
        
        for(x = 0; x < tw; ++x, ++tex, data += 256*3)
         {
          glBindTexture(GL_TEXTURE_2D, *tex);
          
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
                       x+1 == tw ? xrem : 256,
                       y+1 == th ? yrem : 256,
                       0, GL_RGB, GL_UNSIGNED_BYTE,
                       data);
          
          glBegin(GL_QUADS);
          
          glTexCoord2f(0.0f, 0.0f);
          glVertex2f(x*256.0, y*256.0);
          
          glTexCoord2f(0.0f, 1.0f);
          glVertex2f(x*256.0, y*256.0+256.0);
          
          glTexCoord2f(1.0f, 1.0f);
          glVertex2f(x*256.0+256.0, y*256.0+256.0);
          
          glTexCoord2f(1.0f, 0.0f);
          glVertex2f(x*256.0+256.0, y*256.0);
          
          glEnd();
         }
       }
     }
    else
     {
      for(y = 0; y < th; ++y)
       for(x = 0; x < tw; ++x, ++tex)
        {
         glBindTexture(GL_TEXTURE_2D, *tex);
         
         const Region *r = canvas->dirt.start, *e = canvas->dirt.end;
         Region result;
         
         region.x1 = x*256;
         region.y1 = y*256;
         region.x2 = region.x1 + (x+1 == tw ? xrem : 256);
         region.y2 = region.y1 + (y+1 == th ? yrem : 256);
         
         for(; r != e; ++r)
          {
           result = region;
           if(regionClip(&result, r))
            {
             assert(result.x1-region.x1 >= 0);
             assert(result.y1-region.y1 >= 0);
             assert(result.x1-region.x1+result.x2-result.x1 <= region.x2 - region.x1);
             assert(result.y1-region.y1+result.y2-result.y1 <= region.y2 - region.y1);
             
             glTexSubImage2D(GL_TEXTURE_2D, 0,
                             result.x1-region.x1, result.y1-region.y1,
                             result.x2-result.x1, result.y2-result.y1,
                             GL_RGB, GL_UNSIGNED_BYTE,
                             canvas_data+result.x1*3+result.y1*pitch);
            }
          }
         
         glBegin(GL_QUADS);
         
         glTexCoord2f(0.0f, 0.0f);
         glVertex2f(x*256.0, y*256.0);
         
         glTexCoord2f(0.0f, 1.0f);
         glVertex2f(x*256.0, y*256.0+256.0);
         
         glTexCoord2f(1.0f, 1.0f);
         glVertex2f(x*256.0+256.0, y*256.0+256.0);
         
         glTexCoord2f(1.0f, 0.0f);
         glVertex2f(x*256.0+256.0, y*256.0);
         
         glEnd();
        }
     }
    
    DRMakeClean((DirtyRectangles *)&canvas->dirt);
    glPopClientAttrib();
   }
  else
   {
    for(y = 0; y < glc->th; ++y)
     for(x = 0; x < glc->tw; ++x)
      {
       glBindTexture(GL_TEXTURE_2D, *tex);
       
       glBegin(GL_QUADS);
       
       glTexCoord2f(0.0f, 0.0f);
       glVertex2f(x*256.0, y*256.0);
       
       glTexCoord2f(0.0f, 1.0f);
       glVertex2f(x*256.0, y*256.0+256.0);
       
       glTexCoord2f(1.0f, 1.0f);
       glVertex2f(x*256.0+256.0, y*256.0+256.0);
       
       glTexCoord2f(1.0f, 0.0f);
       glVertex2f(x*256.0+256.0, y*256.0);
       
       glEnd();
       
       ++tex;
      }
   }
  
  glPopAttrib();
 }
/*
// Will return 0 if the pixel doesn't exist.
inline unsigned char * getGLCanvasWritePixel(GLCanvas *canvas, signed int x, signed int y)
 {
  x += canvas->xoff;
  y += canvas->yoff;
  
  if(x < 0 || y < 0) return 0;
  
  unsigned int tx = (x>>8), ty = (y>>8);
  
  if(tx >= canvas->columns || ty >= canvas->rows) return 0;
  
  canvas->dirt[tx+ty*canvas->columns] = 1;
  
  return canvas->data
          +(x&0xff)*3
          +(y&0xff)*(3*256)
          +tx*(3*256*256)
          +ty*canvas->tex_pitch;
 }

// Identical to getGLCanvasWritePixel, except your parameters aren't checked for validity.
inline unsigned char * getGLCanvasWritePixelUnsafe(GLCanvas *canvas, signed int x, signed int y)
 {
  x += canvas->xoff;
  y += canvas->yoff;
  
  assert(x >= 0 && y >= 0);
  
  unsigned int tx = (x>>8), ty = (y>>8);
  
  assert(tx < canvas->columns && ty < canvas->rows);
  
  canvas->dirt[tx+ty*canvas->columns] = 1;
  
  return canvas->data
          +(x&0xff)*3
          +(y&0xff)*(3*256)
          +tx*(3*256*256)
          +ty*canvas->tex_pitch;
 }

// Will return 0 if the pixel doesn't exist.
inline const unsigned char * getGLCanvasReadPixel(const GLCanvas *canvas, signed int x, signed int y)
 {
  x += canvas->xoff;
  y += canvas->yoff;
  
  if(x < 0 || y < 0) return 0;
  
  unsigned int tx = (x>>8), ty = (y>>8);
  
  if(tx > canvas->columns || ty > canvas->rows) return 0;
  
  return canvas->data
          +(x&0xff)*3
          +(y&0xff)*(3*256)
          +tx*(3*256*256)
          +ty*canvas->tex_pitch;
 }

// Identical to getGLCanvasReadPixel, except your parameters aren't checked for validity.
inline const unsigned char * getGLCanvasReadPixelUnsafe(const GLCanvas *canvas, signed int x, signed int y)
 {
  x += canvas->xoff;
  y += canvas->yoff;
  
  assert(x >= 0 && y >= 0);
  
  unsigned int tx = (x>>8), ty = (y>>8);
  
  assert(tx < canvas->columns && ty < canvas->rows);
  
  return canvas->data
          +(x&0xff)*3
          +(y&0xff)*(3*256)
          +tx*(3*256*256)
          +ty*canvas->tex_pitch;
 }

// Won't do anything if the pixel doesn't exist.
inline void setGLCanvasPixel(GLCanvas *canvas, signed int x, signed int y,
                             unsigned char r, unsigned char g, unsigned char b)
 {
  unsigned char *pixel = getGLCanvasWritePixel(canvas, x, y);
  
  if(pixel)
   {
    *pixel = r;
    *++pixel = g;
    *++pixel = b;
   }
 }

// Won't do anything if the pixel doesn't exist.
inline void setGLCanvasPixelAlpha(GLCanvas *canvas, signed int x, signed int y,
                                  unsigned char r, unsigned char g, unsigned char b, unsigned char a)
 {
  if(a)
   {
    unsigned char ia = ~a;
    
    unsigned char *pixel = getGLCanvasWritePixel(canvas, x, y);
    
    if(pixel)
     {
      *pixel = (*pixel*ia+r*a)/255;
      ++pixel;
      *pixel = (*pixel*ia+g*a)/255;
      ++pixel;
      *pixel = (*pixel*ia+b*a)/255;
     }
   }
 }

// Identical to setGLCanvasPixel, except your parameters aren't checked for validity.
inline void setGLCanvasPixelUnsafe(GLCanvas *canvas, signed int x, signed int y,
                                   unsigned char r, unsigned char g, unsigned char b)
 {
  unsigned char *pixel = getGLCanvasWritePixelUnsafe(canvas, x, y);
  *pixel = r;
  *++pixel = g;
  *++pixel = b;
 }

// Identical to setGLCanvasPixelAlpha, except your parameters aren't checked for validity.
inline void setGLCanvasPixelAlphaUnsafe(GLCanvas *canvas, signed int x, signed int y,
                                        unsigned char r, unsigned char g, unsigned char b, unsigned char a)
 {
  unsigned char ia = ~a;
  unsigned char *pixel = getGLCanvasWritePixelUnsafe(canvas, x, y);
  
  if(pixel)
   {
    *pixel = (*pixel*ia+r*a)/255;
    ++pixel;
    *pixel = (*pixel*ia+g*a)/255;
    ++pixel;
    *pixel = (*pixel*ia+b*a)/255;
   }
 }

inline void drawGLCanvasPointUnsafe(GLCanvas *canvas, double x, double y,
                                    unsigned char r, unsigned char g, unsigned char b, unsigned char a)
 {
  int ix = floor(x), iy = floor(y);
  float fx = x-ix, ifx = 1.0-fx, fy = y-iy, ify = 1.0-fy;
  
  ix += canvas->xoff;
  iy += canvas->yoff;
  
  assert(ix >= 0 && iy >= 0);
  
  unsigned int tx = (ix>>8), ty = (iy>>8);
  
  assert(tx < canvas->columns && ty < canvas->rows);
  
  ix &= 0xff;
  iy &= 0xff;
  
  unsigned char *dirt = canvas->dirt+(tx+ty*canvas->columns);
  unsigned char *pixel1 = canvas->data+ix*3+iy*(3*256)+tx*(3*256*256)+ty*canvas->tex_pitch,
                *pixel2, *pixel3, *pixel4;
  
  *dirt = 1;
  
  if(ix == 0xff)
   {
    assert(tx+1 != canvas->columns);
    pixel2 = pixel1+256*256*3-255*3;
    
    if(iy == 0xff)
     {
      assert(ty+1 != canvas->rows);
      
      dirt[1] = 1;
      dirt[canvas->columns] = 1;
      dirt[canvas->columns+1] = 1;
      
      signed int o = canvas->tex_pitch-255*256*3;
      pixel3=pixel1+o;
      pixel4=pixel2+o;
     }
    else
     {
      dirt[1] = 1;
      pixel3=pixel1+3*256;
      pixel4=pixel2+3*256;
     }
   }
  else
   {
    pixel2 = pixel1+3;
    
    if(iy == 0xff)
     {
      dirt[canvas->columns] = 1;
      assert(ty+1 != canvas->rows);
      signed int o = canvas->tex_pitch-255*256*3;
      pixel3=pixel1+o;
      pixel4=pixel2+o;
     }
    else
     {
      pixel3=pixel1+3*256;
      pixel4=pixel2+3*256;
     }
   }
  
  unsigned char na = (ifx*ify*a+0.5), ia = ~na;
  *pixel1 = (*pixel1*ia+r*na)/255; ++pixel1;
  *pixel1 = (*pixel1*ia+g*na)/255; ++pixel1;
  *pixel1 = (*pixel1*ia+b*na)/255;
  
  na = (fx*ify*a+0.5); ia = ~na;
  *pixel2 = (*pixel2*ia+r*na)/255; ++pixel2;
  *pixel2 = (*pixel2*ia+g*na)/255; ++pixel2;
  *pixel2 = (*pixel2*ia+b*na)/255;
  
  na = (ifx*fy*a+0.5); ia = ~na;
  *pixel3 = (*pixel3*ia+r*na)/255; ++pixel3;
  *pixel3 = (*pixel3*ia+g*na)/255; ++pixel3;
  *pixel3 = (*pixel3*ia+b*na)/255;
  
  na = (fx*fy*a+0.5); ia = ~na;
  *pixel4 = (*pixel4*ia+r*na)/255; ++pixel4;
  *pixel4 = (*pixel4*ia+g*na)/255; ++pixel4;
  *pixel4 = (*pixel4*ia+b*na)/255;
 }
*/

void renderGLCanvasMandelbrot(Canvas *canvas)
 {
  float p2x = 2.0/(canvas->height);
  float p2y = 2.0/(canvas->height);
  float ox = (2.0/(canvas->width))*0.5-(2.0/(canvas->height))*0.5-2;
  float oy = -1;
  
  unsigned int y, x;
  
  for(y = 0; y < canvas->height; ++y)
   for(x = 0; x < canvas->width; ++x)
    {
     float fx = x*p2x+ox, cx=fx, fy=y*p2y+oy, cy=fy;
     
     for(unsigned int i = 0; i < 512; i+=2)
      {
       if(fx*fx+fy*fy > 4.0)
        {
         float t = fx*fx-fy*fy+cx;
         fy = 2*fx*fy+cy;
         fx = t*t-fy*fy+cx;
         fy = 2*t*fy+cy;
         t = fx*fx-fy*fy+cx;
         fy = 2*fx*fy+cy;
         
         float p = i+4-log(log(sqrt(t*t+fy*fy)))/log(2);
         
         /*setGLCanvasPixelUnsafe(canvas, x, y, sin(p)*127.5+127.5,
                                              sin(p+2.0943951023931954923084289221863352561314463)*127.5+127.5,
                                              sin(p+4.1887902047863909846168578443726705122628925)*127.5+127.5);
         */
         setPixelUnsafe(canvas, x, y, sin(p)*127.5+127.5,
                        sin(p+2.0943951023931954923084289221863352561314463)*127.5+127.5,
                        sin(p+4.1887902047863909846168578443726705122628925)*127.5+127.5);
         
         goto done; // Yes, yes, goto's are evil. Pretend it is a continue, for the loop outside this one.
        }
       
       float t = fx*fx-fy*fy+cx;
       fy = 2*fx*fy+cy;
       
       if(t*t+fy*fy > 4.0)
        {
         fx = t*t-fy*fy+cx;
         fy = 2*t*fy+cy;
         t = fx*fx-fy*fy+cx;
         fy = 2*fx*fy+cy;
         fx = t*t-fy*fy+cx;
         fy = 2*t*fy+cy;
         
         float p = i+5-log(log(sqrt(fx*fx+fy*fy)))/log(2);
         
         /*setGLCanvasPixelUnsafe(canvas, x, y, sin(p)*127.5+127.5,
                                        sin(p+2.0943951023931954923084289221863352561314463)*127.5+127.5,
                                        sin(p+4.1887902047863909846168578443726705122628925)*127.5+127.5);*/
         
         setPixelUnsafe(canvas, x, y, sin(p)*127.5+127.5,
                        sin(p+2.0943951023931954923084289221863352561314463)*127.5+127.5,
                        sin(p+4.1887902047863909846168578443726705122628925)*127.5+127.5);
         
         goto done;
        }
       
       fx = t*t-fy*fy+cx;
       fy = 2*t*fy+cy;
      }
     
     setPixelUnsafe(canvas, x, y, 0, 0, 0);
     
     done: continue;
    }
 }


