#include "common.h"
#include "region.h"

/* Region functions. */

int regionCollision(const Region * const a, const Region * const b)
 {
  assert(a);
  assert(b);
  assert(regionIsNormal(a));
  assert(regionIsNormal(b));
  
  return a->x1 < b->x2 &&
         a->y1 < b->y2 &&
         a->x2 > b->x1 &&
         a->y2 > b->y1;
 }

int regionIsNormal(const Region * const r)
 {
  assert(r);
  
  return r->x2 >= r->x1 && r->y2 >= r->y1;
 }

void regionNormalize(Region * const r)
 {
  assert(r);
  
  if(r->x1 > r->x2)
   {
    int temp = r->x1;
    r->x1 = r->x2 - 1;
    r->x2 = temp + 1;
   }
  
  if(r->y1 > r->y2)
   {
    int temp = r->y1;
    r->y1 = r->y2 - 1;
    r->y2 = temp + 1;
   }
 }

int regionClip(Region * const r, const Region * const base)
 {
  assert(r);
  assert(regionIsNormal(r));
  assert(base);
  assert(regionIsNormal(base));
  
  if(r->x1 == r->x2 || r->y1 == r->y2 || base->x1 == base->x2 || base->y1 == base->y2)
   return 0; /* Base either has no volume and nothing can fit in it, or region has no volume and therefore doesn't exist. */
  
  if(r->x1 < base->x1 && (r->x1 = base->x1) >= r->x2)
   return 0;
  
  if(r->y1 < base->y1 && (r->y1 = base->y1) >= r->y2)
   return 0;
  
  if(r->x2 > base->x2 && (r->x2 = base->x2) <= r->x1)
   return 0;
  
  if(r->y2 > base->y2 && (r->y2 = base->y2) <= r->y1)
   return 0;
  
  return 1;
 }




/* DirtyRectangles functions. */

void DRInit(DirtyRectangles * const dr)
 {
  assert(dr);
  
  dr->end = dr->start;
 }


static int DRAddHelper(const Region *i, DirtyRectangles * const dr, const Region * const r)
 {
  assert(dr);
  assert(r);
  assert(regionIsNormal(r));
  
  if(r->x1 == r->x2 || r->y1 == r->y2)
   return 1; /* It has no volume, so no need to add it. */
  
  if(dr->end >= dr->start + sizeof dr->start / sizeof *dr->start)
   { /* Adding another element would saturate the list. */
    DRMakeSaturated(dr);
    return 0;
   }
  
  for(; i != dr->end; ++i)
   if(regionCollision(i, r))
    {
     Region temp = *r; /* We're going to make some changes, so we'll need a place to store the copy. */
     
     if(r->y1 < i->y1)
      {
       temp.y2 = i->y1;
       
       assert(!regionCollision(i, &temp));
       
       if(!DRAddHelper(i+1, dr, &temp))
        return 0;
       
       temp.y2 = r->y2;
      }
     
     if(r->y2 > i->y2)
      {
       
       temp.y1 = i->y2;
       
       assert(!regionCollision(i, &temp));
       
       if(!DRAddHelper(i+1, dr, &temp))
        return 0;
      }
     
     if(r->y1 < i->y1)
      temp.y1 = i->y1;
     else
      temp.y1 = r->y1;
     
     if(r->y2 > i->y2)
      temp.y2 = i->y2;
     else
      temp.y2 = r->y2;
     
     if(r->x1 < i->x1)
      {
       temp.x2 = i->x1;
       
       assert(!regionCollision(i, &temp));
       
       if(!DRAddHelper(i+1, dr, &temp))
        return 0;
       
       temp.x2 = r->x2;
      }
     
     if(r->x2 > i->x2)
      {
       
       temp.x1 = i->x2;
       
       assert(!regionCollision(i, &temp));
       
       if(!DRAddHelper(i+1, dr, &temp))
        return 0;
      }
     
     return 1;
    }
  
  *dr->end++ = *r;
  return 1;
 }

int DRAdd(DirtyRectangles * const dr, const Region * const r)
 {
  return DRAddHelper(dr->start, dr, r);
 }

static int DRRemoveHelper(DirtyRectangles * const dr, const Region * const r, Region * const i)
 {
  Region temp, v;
  
  assert(dr);
  assert(r);
  assert(i);
  assert(regionIsNormal(r));
  assert(regionIsNormal(i));
  assert(i->x1 < i->x2 && i->y1 < i->y2);
  assert(regionCollision(r, i));
  
  temp = v = *i;
  *i = *(--dr->end);
  
  v.y1 = r->y2;
  if(regionIsNormal(&v) && !DRAdd(dr, &v)) return 0;
  v.y1 = temp.y1;
  
  v.y2 = r->y1;
  if(regionIsNormal(&v) && !DRAdd(dr, &v)) return 0;
  v.y2 = temp.y2;
  
  if(v.y1 < r->y1) temp.y1 = v.y1 = r->y1;
  if(v.y2 > r->y2) temp.y2 = v.y2 = r->y2;
  
  temp.x1 = r->x2;
  if(regionIsNormal(&temp) && !DRAdd(dr, &temp)) return 0;
  temp.x1 = v.x1;
  
  temp.x2 = r->x1;
  if(regionIsNormal(&temp) && !DRAdd(dr, &temp)) return 0;
  
  return 1;
 }

int DRRemove(DirtyRectangles * const dr, const Region * const r)
 {
  Region *i;
  
  assert(dr);
  assert(r);
  assert(regionIsNormal(r));
  
  if(r->x1 == r->x2 || r->y1 == r->y2)
   return 1; /* It has no volume, so no need to remove it. */
  
  if(DRIsSaturated(dr))
   {
    /* Not sure how to handle this. For now, we'll just do nothing and leave it saturated. */
    return 0;
   }
  
  for(i = dr->start; i != dr->end; ++i)
   if(regionCollision(i, r))
    {
     if(DRRemoveHelper(dr, r, i))
      --i;
     else
      return 0;
    }
   
  return 1;
 }

int DRAddGroup(DirtyRectangles * const dr, const int16_t x, const int16_t y, const DirtyRectangles * const regions)
 {
  const Region *r;
  
  assert(dr);
  assert(regions);
  
  if(dr == regions)
   return 1; /* That was simple enough. */
  
  if(DRIsSaturated(regions))
   {
    DRMakeSaturated(dr);
    return 0;
   }
  
  for(r = regions->start; r != regions->end; ++r)
   {
    Region temp;
    temp.x1 = x+r->x1;
    temp.y1 = y+r->y1;
    temp.x2 = x+r->x2;
    temp.y2 = y+r->y2;
    
    if(!DRAdd(dr, &temp))
     return 0;
   }
  
  return 1;
 }

int DRRemoveGroup(DirtyRectangles * const dr, const int16_t x, const int16_t y, const DirtyRectangles * const regions)
 {
  assert(dr);
  assert(regions);
  
  if(DRIsClean(dr))
   { /* Assuming there is actually something to remove from. . . */
    if(dr == regions || DRIsSaturated(regions))
     DRMakeClean(dr); /* Everything gets removed. Joy. */
    else
     {
      const Region *r = regions->start;
      
      for(; r != regions->end; ++r)
       {
        Region temp;
        temp.x1 = x+r->x1;
        temp.y1 = y+r->y1;
        temp.x2 = x+r->x2;
        temp.y2 = y+r->y2;
        
        if(!DRRemove(dr, &temp))
         return 0;
       }
     }
   }
  
  return 1;
 }

int DRIsSaturated(const DirtyRectangles * const dr)
 {
  assert(dr);
  
  return dr->end == dr->start + sizeof dr->start / sizeof *dr->start + 1;
 }

int DRIsDirty(const DirtyRectangles * const dr)
 {
  return dr->end != dr->start;
 }

extern int DRIsClean(const DirtyRectangles * const dr)
 {
  return dr->end == dr->start;
 }

void DRMakeSaturated(DirtyRectangles * const dr)
 {
  assert(dr);
  
  dr->end = dr->start + sizeof dr->start / sizeof *dr->start + 1;
 }

void DRMakeClean(DirtyRectangles * const dr)
 {
  assert(dr);
  
  dr->end = dr->start;
 }

void DRClip(DirtyRectangles * const dr, const Region * const r)
 {
  Region *i;
  
  assert(dr);
  assert(r);
  assert(regionIsNormal(r));
  
  if(DRIsSaturated(dr))
   {
    /* We'll take the shape of the region. */
    *dr->start = *r;
    dr->end = dr->start + 1;
    return;
   }
  
  for(i = dr->start; i != dr->end; ++i)
   if(!regionClip(i, r))
    *i-- = *(--dr->end);
 }

static int DROptimizeHelper(DirtyRectangles * const dr)
 {
  char changed = 0;
  Region *r1, *r2;
  
  assert(dr);
  
  for(r1 = dr->start; r1 < dr->end - 1; ++r1)
   for(r2 = r1+1; r2 < dr->end; ++r2)
    {
     assert(r1->x1 < r1->x2 && r1->y1 < r1->y2);
     assert(r2->x1 < r2->x2 && r2->y1 < r2->y2);
     assert(!regionCollision(r1, r2));
     
     if(r1->x1 == r2->x1 && r1->x2 == r2->x2)
      {
       if(r1->y2 == r2->y1)
        {
         r1->y2 = r2->y2;
         goto remove_r2;
        }
       else if(r1->y1 == r2->y2)
        {
         r1->y1 = r2->y1;
         goto remove_r2;
        }
      }
     
     if(r1->y1 == r2->y1 && r1->y2 == r2->y2)
      {
       if(r1->x2 == r2->x1)
        {
         r1->x2 = r2->x2;
         goto remove_r2;
        }
       else if(r1->x1 == r2->x2)
        {
         r1->x1 = r2->x1;
         goto remove_r2;
        }
      }
     
     if((r1->x2 == r2->x1 || r1->x1 == r2->x2) && r1->y1 < r2->y2 && r1->y2 > r2->y1)
      { /* This will attempt to split and extend rectanges horizontally. Since this means making
           possibly more rectangles than we can hold, we need to be careful. */
       Region start[3], *end = start;
       
       /* This is a helper to help us figure out what we need. */
       /* TODO: I'm pretty sure some of them we can know for sure if they will exist.
          When I get really bored and feel like doing useless optimizations, figure
          out which ones to remove the checks for. */
       #define ADD(A, B, C, D) if((A)<(C)&&(B)<(D)){end->x1=(A);end->y1=(B);end->x2=(C);end->y2=(D);++end;}
       
       if(r1->x2 == r2->x1)
        {
         if(r1->y1 < r2->y1 && r1->y2 > r2->y2)
          {
           ADD(r1->x1, r1->y1, r1->x2, r2->y1); /* +--+    */
           ADD(r1->x1, r2->y1, r2->x2, r2->y2); /* |  +--+ */
           ADD(r1->x1, r2->y2, r1->x2, r1->y2); /* +--+    */
          }
         else if(r1->y1 > r2->y1 && r1->y2 < r2->y2)
          {
           ADD(r1->x2, r2->y1, r2->x2, r1->y1); /*    +--+ */
           ADD(r1->x1, r1->y1, r2->x2, r1->y2); /* +--+  | */
           ADD(r1->x2, r1->y2, r2->x2, r2->y2); /*    +--+ */
          }
         else if(r1->y1 <= r2->y1 && r1->y2 <= r2->y2)
          {
           ADD(r1->x1, r1->y1, r1->x2, r2->y1); /* +--+    */
           ADD(r1->x1, r2->y1, r2->x2, r1->y2); /* +--+--+ */
           ADD(r1->x2, r1->y2, r2->x2, r2->y2); /*    +--+ */
          }
         else
          {
           ADD(r1->x2, r2->y1, r2->x2, r1->y1); /*    +--+ */
           ADD(r1->x1, r1->y1, r2->x2, r2->y2); /* +--+--+ */
           ADD(r1->x1, r2->y2, r1->x2, r1->y2); /* +--+    */
          }
        }
       else
        { /* Same as above with r1 and r2 swapped. Mirror any changes above here. */
         assert(r1->x1 == r2->x2);
         
         if(r2->y1 < r1->y1 && r2->y2 > r1->y2)
          {
           ADD(r2->x1, r2->y1, r2->x2, r1->y1);
           ADD(r2->x1, r1->y1, r1->x2, r1->y2);
           ADD(r2->x1, r1->y2, r2->x2, r2->y2);
          }
         else if(r2->y1 > r1->y1 && r2->y2 < r1->y2)
          {
           ADD(r2->x2, r1->y1, r1->x2, r2->y1);
           ADD(r2->x1, r2->y1, r1->x2, r2->y2);
           ADD(r2->x2, r2->y2, r1->x2, r1->y2);
          }
         else if(r2->y1 <= r1->y1 && r2->y2 <= r1->y2)
          {
           ADD(r2->x1, r2->y1, r2->x2, r1->y1);
           ADD(r2->x1, r1->y1, r1->x2, r2->y2);
           ADD(r2->x2, r2->y2, r1->x2, r1->y2);
          }
         else
          {
           ADD(r2->x2, r1->y1, r1->x2, r2->y1);
           ADD(r2->x1, r2->y1, r1->x2, r1->y2);
           ADD(r2->x1, r1->y2, r2->x2, r2->y2);
          }
        }
       #undef ADD
       
       if(end != start)
        {
         /*
         logMessage(Notice, "-------------");
         logMessage(Notice, "[%d, %d] - [%d, %d]", r1->x1, r1->y1, r1->x2, r1->y2);
         logMessage(Notice, "[%d, %d] - [%d, %d]", r2->x1, r2->y1, r2->x2, r2->y2);
         logMessage(Notice, "----- v -----");
         Region *p = start;
         for(; p != end; ++p)
          logMessage(Notice, "[%d, %d] - [%d, %d]", p->x1, p->y1, p->x2, p->y2);
         logMessage(Notice, "----- v -----");
         */
         assert(end - start > 1 && end-start < 4);
         
         if(dr->end+(end-start) <= dr->start + sizeof dr->start / sizeof *dr->start)
          { /* Now, add these new rectangles. */
           *r1 = start[0];
           
           /*if(end == start+1)
            goto remove_r2; // No idea how this could happen. WTF. . .*/
           
           *r2 = start[1];
           
           if(end == start+3)
            DRAdd(dr, start+2);
           
           assert(!DRIsSaturated(dr));
           
           --r1;
           changed = 1;
           break;
          }
        }
      }
     
     continue;
     
     remove_r2:
     *r2-- = *(--dr->end);
     changed = 1;
    }
  
  return changed;
 }

void DROptimize(DirtyRectangles * const dr)
 {
  assert(dr);
  
  if(!DRIsSaturated(dr))
   while(DROptimizeHelper(dr))
    { /* Keep going until nothing more can be done. */ }
 }

